/*
 * tcc.c - TC compiler
 *
 * Written 2001 by Werner Almesberger
 * Copyright 2001 EPFL-ICA, Network Robots
 */


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>

#include "config.h"
#include "error.h"
#include "util.h"
#include "data.h"
#include "var.h"
#include "tree.h"
#include "device.h"
#include "target.h"
#include "if.h"
#include "ext.h"
#include "ext_all.h"
#include "location.h"


#define STRING(s) #s


extern int yyparse(void);


DATA_LIST *pragma = NULL;
const char *default_device = DEFAULT_DEVICE;
const char *location_file = NULL;
const char *var_use_file = NULL;
int alg_mode = 0;
int remove_qdiscs = 0;
int quiet = 0;
int debug = 0;
int no_warn = 0;
int cheap_only = 0;


/* ----- CPP child process ------------------------------------------------- */


static pid_t pid;


static void kill_cpp(void)
{
    (void) kill(pid,SIGTERM);
}


static void run_cpp(int argc,char *const *argv)
{
    int fds[2];

    if (pipe(fds) < 0) {
	perror("pipe");
	exit(1);
    }
    pid = fork();
    if (pid < 0) {
	perror("fork");
	exit(1);
    }
    if (!pid) {
	if (dup2(fds[1],1) < 0) {
	    perror("dup2");
	    exit(1);
	}
	if (execvp(CPP,argv) < 0) {
	    perror("execvp " CPP);
	    exit(1);
	}
	/* not reached */
    }
    if (dup2(fds[0],0) < 0) {
	perror("dup2");
	exit(1);
    }
    if (close(fds[1]) < 0) {
	perror("close");
	exit(1);
    }
    atexit(kill_cpp);
}


/* ----- Post preamble processing ------------------------------------------ */


static struct warn_switch  {
    const char *name;
    struct warn_switch *next;
} *warn_switches = NULL,**last_warn_switch = &warn_switches;


static void record_warn_switch(const char *name)
{
    *last_warn_switch = alloc_t(struct warn_switch);
    (*last_warn_switch)->name = stralloc(name);
    (*last_warn_switch)->next = NULL;
    last_warn_switch = &(*last_warn_switch)->next;
}


void set_warn_switches(void); /* tcng.y uses extern */
void set_warn_switches(void)
{
    const  struct warn_switch *walk;

    for (walk = warn_switches; walk; walk = walk->next)
	if (!set_warn_switch(warn_switches->name)) abort();
}


static void free_warn_switches(void)
{
    while (warn_switches) {
	struct warn_switch *next;

	next = warn_switches->next;
	free((char *) warn_switches->name);
	free(warn_switches);
	warn_switches = next;
    }
}


/* ----- Command-line parsing ---------------------------------------------- */


static void usage(const char *name)
{
    fprintf(stderr,"usage: %s [-c] [-d ...] [-E] [-i default_interface]\n",
      name);
    fprintf(stderr,"%10s [-l location_file] [-q] [-r] [-w] [-W[no]...]\n","");
    fprintf(stderr,"%10s [-x elem:ext_target ...] [-t [no]target ...]\n","");
    fprintf(stderr,"%10s [-Xphase,arg] [cpp_option ...] [file]\n","");
    fprintf(stderr,"%6s %s -V\n\n","",name);
    fprintf(stderr,"  -c                    only check validity\n");
    fprintf(stderr,"  -d ...                increase debugging level\n");
    fprintf(stderr,"  -E                    refuse \"expensive\" constructs in"
      " classifier\n");
    fprintf(stderr,"  -i default_interface  interface if not specified in "
      "description file\n");
    fprintf(stderr,"                        (default: "
      STRING(DEFAULT_DEVICE) ")\n");
    fprintf(stderr,"  -l location_file      write a list of source code "
      "locations of traffic\n");
    fprintf(stderr,"                        control elements (\"stderr\" for "
      "standard error)\n");
    fprintf(stderr,"  -q                    quiet, produce terse output\n");
    fprintf(stderr,"  -r                    remove old qdiscs\n");
    fprintf(stderr,"  -t [elem:][no]target  enable or disable target\n");
    fprintf(stderr,"                        elements: if\n");
    fprintf(stderr,"                        targets: all, tc, c, ext; default: "
      "tc\n");
    fprintf(stderr,"  -u variable_use_file  for each variable, write name and "
      "value\n");
    fprintf(stderr,"  -V                    print version number and exit\n");
    fprintf(stderr,"  -w                    suppress all warnings\n");
    fprintf(stderr,"  -W[no]expensive       like -E, but warn only\n");
    fprintf(stderr,"  -W[no]redefine        warn if re-defining variables\n");
    fprintf(stderr,"  -W[no]truncate        warn if truncating values\n");
    fprintf(stderr,"  -W[no]unused          report unused variables\n");
    fprintf(stderr,"  -x elem:ext_target    register external target\n");
    fprintf(stderr,"  -Xphase,arg           verbatim argument for specific "
      "build phase\n");
    fprintf(stderr,"                        phases: p=cpp, m=tcc-module, "
      "k=kmod_cc, t=tcmod_cc,\n");
    fprintf(stderr,"                                x=external (all), xelem="
      "external (element)\n");
    fprintf(stderr,"  cpp_option            -Idir, -Dname[=value], or "
      "-Uname\n");
    exit(1);
}


int main(int argc,char *const *argv)
{
    const char *tcng_topdir;
    char *include;
    char opt[3] = "-?";
    char **cpp_argv;
    int check_only = 0;
    int c,cpp_argc = 1;

    target_register("if","tc",1);
    target_register("if","c",0);
    target_register("if","ext",0);
    cpp_argv = alloc(sizeof(char *)*(argc*2+1)); /* generous allocation */
    tcng_topdir = getenv("TCNG_TOPDIR");
    include = alloc_sprintf("-I%s/" TCNG_INC_DIR,
      tcng_topdir ? tcng_topdir : TOPDIR);
    cpp_argv[cpp_argc++] = include;
    /*
     * Reserved options (strong Unix tradition):
     *
     * -o file  for some output
     * -f file  or -f for "force"
     * -v       verbose
     */
    while ((c = getopt(argc,argv,"BNcdEhi:l:qrt:u:W:wx:D:U:I:VX:")) != EOF)
	switch (c) {
	    case 'c':
		check_only = 1;
		break;
	    case 'B':
		use_bit_tree = 1;
		break;
	    case 'N':
		alg_mode++;
		break;
	    case 'd':
		debug++;
		break;
	    case 'E':
		cheap_only = 1;
		break;
	    case 'h':
		usage(argv[0]);
	    case 'i':
		default_device = optarg;
		break;
	    case 'l':
		location_file = optarg;
		break;
	    case 'q':
		quiet = 1;
		break;
	    case 'r':
		remove_qdiscs = 1;
		break;
	    case 't':
		switch_target(optarg);
		break;
	    case 'u':
		var_use_file = optarg;
		break;
	    case 'V':
		printf("tcc (tcng) version %s\n",VERSION);
		return 0;
	    case 'W':
		if (!set_warn_switch(optarg)) usage(argv[0]);
		record_warn_switch(optarg);
		break;
	    case 'w':
		no_warn = 1;
		break;
	    case 'x':
		ext_target_register(optarg);
		break;
            case 'X':
		if (*optarg == 'p')  {
		    cpp_argv[cpp_argc++] = stralloc(optarg+2);
		    break;
		}
		if (*optarg == 'x') {
		    char *tmp = stralloc(optarg);
		    char *comma;

		    comma = strchr(tmp,',');
		    if (!comma) usage(argv[0]);
		    *comma = 0;
		    add_tcc_external_arg(tmp+1,comma+1);
		    free(tmp);
		    break;
		}
		if (!optarg[0] || !strchr("mkt",optarg[0]) || optarg[1] != ',')
		    usage(argv[0]);
		add_tcc_module_arg(optarg[0] == 'm',optarg);
                break;
	    case 'D':
	    case 'U':
	    case 'I':
		opt[1] = c;
		cpp_argv[cpp_argc++] = stralloc(opt);
		cpp_argv[cpp_argc++] = optarg;
		break;
	    default:
		usage(argv[0]);
	}
    ext_targets_configure();
    while (optind < argc) cpp_argv[cpp_argc++] = argv[optind++];
    cpp_argv[0] = argv[0];
    cpp_argv[cpp_argc] = NULL;
    run_cpp(cpp_argc,cpp_argv);
    free(include);
    (void) yyparse();
    free_warn_switches();
    check_devices();
    var_end_scope(); /* report unused variables */
    if (!use_bit_tree) {
	switch (alg_mode) {
	    case 0: /* new: no -N uses the "new" optimization */
	    case 1: /* -N has the expected effect */
		alg_mode = 1;
		break;
	    case 2: /* new: -N -N switches to the old optimization */
		break;
	    default:
		error("too many -N options");
	}
    }
    if (location_file) write_location_map(location_file);
    if (var_use_file) write_var_use(var_use_file);
    if (!check_only) {
	if (dump_all && ext_targets) dump_all_devices(ext_targets->name);
		/* @@@ walk list of targets */
	else dump_devices();
    }
    return 0;
}
